/***********************************************************************************************************************
 *
 * blueMarine Metadata - open source media workflow
 * Copyright (C) 2007-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluemarine.tidalwave.it
 * SCM: https://kenai.com/hg/bluemarine~metadata-src
 *
 **********************************************************************************************************************/
package it.tidalwave.metadata.impl;

import it.tidalwave.metadata.spi.TestLoggingSupport;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.beans.Introspector;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyDescriptor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import it.tidalwave.util.logging.Logger;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.awt.Dimension;
import java.awt.color.ColorSpace;
import java.awt.color.ICC_Profile;
import java.text.SimpleDateFormat;
import org.jdesktop.beansbinding.BeanProperty;
import org.jdesktop.beansbinding.Property;
import static org.junit.Assert.*;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public abstract class EnhancedMetadataItemTestSupport extends TestLoggingSupport
  {
    private final static String CLASS = EnhancedMetadataItemTestSupport.class.getName();
    private final static Logger logger = Logger.getLogger(CLASS);
    
    private final static int ITERATIONS = 10;

    private final static List<String> IGNORED_PROPERTIES = 
            Arrays.asList("class", "propertyChangeListeners", "vetoableChangeListeners", "callback", "callbacks", "latestModificationTime");
    
    class PropertyChangeListenerTester implements PropertyChangeListener
      {
        public final List<String> latestPropertyName = new ArrayList<String>();
        public final List<Object> latestPropertyValue = new ArrayList<Object>();
        
        public void reset()
          {
            latestPropertyName.clear();
            latestPropertyValue.clear();
          }
        
        public void propertyChange (final PropertyChangeEvent event) 
          {
            final String propertyName = event.getPropertyName();
            final Object newValue = event.getNewValue();
            logger.fine("**** changed %s: %s -> %s", propertyName, toStringx(event.getOldValue()), toStringx(newValue));
            
            if (!Arrays.asList("available", "latestModificationTime").contains(propertyName))
              {
//        if (bean.getClass().getName().indexOf("EXIF") >= 0) // FIXME: workarounds for EXIF

//            if (!propertyName.endsWith("AsDate"))
//              {
                latestPropertyName.add(propertyName);
                latestPropertyValue.add(newValue);
//              }
                
              }
          }
      }
            
    protected void testAllProperties() 
      throws Exception
      {
        final Class<?> beanClass = createBean().getClass();
        
        for (final PropertyDescriptor propertyDescriptor : Introspector.getBeanInfo(beanClass).getPropertyDescriptors())
          {
            if (!IGNORED_PROPERTIES.contains(propertyDescriptor.getName()))
              {
                testProperty(createBean(), createBean(), propertyDescriptor);
              }
          }
      }
    
    protected abstract Object createBean();
    
    private void testProperty (final Object bean, final Object emptyReference, final PropertyDescriptor propertyDescriptor)
      throws Exception
      {
        String propertyName =  propertyDescriptor.getName();
        
        final Class<?> propertyType = propertyDescriptor.getPropertyType();
        logger.info("******************************** testProperty(%s, %s) - type %s", bean, propertyName, propertyType);
        
        final Property<Object, Object> property = BeanProperty.<Object, Object>create(propertyName);
        final PropertyChangeListenerTester pclt = new PropertyChangeListenerTester();
//        bean.addPropertyChangeListener(pclt);        
        final Method method = bean.getClass().getMethod("addPropertyChangeListener", PropertyChangeListener.class);
        method.invoke(bean, pclt);
        
        if (!property.isWriteable(bean))
          {
            logger.finest(">>>> is read only, skipping");
            return;
          }
        
        if (bean.getClass().getName().indexOf("IPTC") >= 0) // FIXME
          {
            propertyName = propertyName.replaceAll("IPTC", "iptc");
          }

        if (bean.getClass().getName().indexOf("EXIF") >= 0) // FIXME: workarounds for EXIF
          {
            if ("object".equals(propertyName)) // it's not meant to be public - better to replace with something not get/set?
              {
                logger.warning(">>>> skipping");
                return;
              }
            
//            if (propertyName.endsWith("AsDate")) // fires a different property
//              {
////                propertyName = propertyName.replaceAll("AsDate$", "");
//                logger.warning(">>>> skipping");
//                return;
//              }
          }

        if (Collection.class.isAssignableFrom(propertyType))
          {
            if (bean.getClass().getName().indexOf("Basic") < 0) // FIXME
              {
                pclt.reset();
                property.setValue(bean, null);
                assertNull(property.getValue(bean));

                final Collection<String> value2 = Arrays.asList("1", "2");
                pclt.reset();
                property.setValue(bean, value2);

                // Test that you can change the collection contents without affecting the bean
                Collection<Object> collection = (Collection<Object>)property.getValue(bean);
                assertEquals(value2, collection);
                collection.add("3");
                assertEquals(2, value2.size());
                assertEquals(3, collection.size());

                collection = (Collection<Object>)property.getValue(bean);         
                assertEquals(value2, collection);
              } 
          }
        
        else if (!propertyName.endsWith("Available") && !propertyName.equals("available"))
          {
            for (int i = 0; i < ITERATIONS; i++)
              {
                pclt.reset();
                
                final Object previousValue = property.getValue(bean);
                Object value = createValue(i, propertyType, propertyName);
                
                logger.finest("================== current value %s - setting to %s", toStringx(previousValue), toStringx(value));
                property.setValue(bean, value);
                assertEquals(value, property.getValue(bean));
                
                final List<String> extraProperties = new ArrayList<String>();
                
                if (propertyName.startsWith("dateTime") && !propertyName.endsWith("AsDate"))
                  {
                    extraProperties.add(propertyName + "AsDate");  
                    
                    if (i == 0)
                      {
                        extraProperties.add(propertyName + "AsDateAvailable");  
                      }
                  }
                
                if (propertyName.startsWith("dateTime") && propertyName.endsWith("AsDate"))
                  {
                    // TODO: too complex for now, don't test it
                    // [dateTimeAsDate, dateTimeAsDateAvailable, available, latestModificationTime]> 
                    // but was:<[dateTime, dateTimeAvailable, available, latestModificationTime, dateTimeAsDate, dateTimeAsDateAvailable]>
                    continue;
                  }
                
                if (propertyName.equals("userCommentAsString"))
                  {
                    // TODO: too complex for now, don't test it
                    continue;
                  }
                
                final List<String> expectedProperties = new ArrayList<String>();
                
                //
                // propertyAvailable must change state from false to true only at the beginning.
                // Even setting a property to null and after to not-null mustn't change it.
                //
                if (i == 0) 
                  {
//                    expectedProperties.addAll(Arrays.asList(propertyName, propertyName + "Available", "available", "latestModificationTime"));
                    expectedProperties.addAll(Arrays.asList(propertyName, propertyName + "Available"));
                    expectedProperties.addAll(extraProperties);
                    assertEquals(expectedProperties, pclt.latestPropertyName);
                    assertEquals(Arrays.asList(value, true), pclt.latestPropertyValue.subList(0, 2));
//                    assertEquals(Arrays.asList(value, true, true), pclt.latestPropertyValue.subList(0, 3));
                  }
                else 
                  {
                    expectedProperties.addAll(Arrays.asList(propertyName));
//                    expectedProperties.addAll(Arrays.asList(propertyName, "latestModificationTime"));
                    expectedProperties.addAll(extraProperties);
                    assertEquals(expectedProperties, pclt.latestPropertyName);
                    assertEquals(Arrays.asList(value), pclt.latestPropertyValue.subList(0, 1));
                  }
                
                // Always different even when you set the property to null, since its availability stays to true
                assertFalse(emptyReference.equals(bean));
                assertFalse(bean.equals(emptyReference));
                // Since they are different we can't assert anything about hashcodes, just call them for the coverage tool
                emptyReference.hashCode();
                bean.hashCode();
                  
                final Object cloned = clone(bean);
                if ((bean.getClass().getName().indexOf("EXIF") < 0) // FIXME
                    && (bean.getClass().getName().indexOf("TIFF") < 0)
                    && (bean.getClass().getName().indexOf("Basic") < 0))
                  {
                    assertTrue(cloned.equals(bean));
                    assertTrue(bean.equals(cloned));
                    assertEquals(bean.hashCode(), cloned.hashCode());
                  }
              }
            //
            // Setting propertyAvailable to false must change the property value to null
            //
            assertNotNull(property.getValue(bean));
            final Property<Object, Boolean> availabilityProperty = BeanProperty.<Object, Boolean>create(propertyName + "Available");
            pclt.reset();
            final Object previousValue = property.getValue(bean);
            logger.finest(">>>> current value %s - setting availability to false",  toStringx(previousValue));
            availabilityProperty.setValue(bean, false);
            assertNull(property.getValue(bean));
            
            if (propertyName.startsWith("dateTime") && propertyName.endsWith("AsDate"))
              {
                // TODO: too complex for now, don't test it
                // [dateTimeAsDate, dateTimeAsDateAvailable, available, latestModificationTime]> 
                // but was:<[dateTime, dateTimeAvailable, available, latestModificationTime, dateTimeAsDate, dateTimeAsDateAvailable]>
                return;
              }
            
            if (propertyName.equals("userCommentAsString"))
              {
                // TODO: too complex for now, don't test it
                return;
              }
            
//            assertEquals(Arrays.asList(propertyName + "Available", propertyName, "available", "latestModificationTime"), pclt.latestPropertyName);
//            assertEquals(Arrays.asList(false, null, false), pclt.latestPropertyValue.subList(0, 3));
            assertEquals(Arrays.asList(propertyName + "Available", propertyName), pclt.latestPropertyName);
            assertEquals(Arrays.asList(false, null), pclt.latestPropertyValue.subList(0, 2));
            
            // Now that availability has been reset to false, the object must be equal to the original
            if ((bean.getClass().getName().indexOf("EXIF") < 0) // FIXME
                && (bean.getClass().getName().indexOf("TIFF") < 0)
                && (bean.getClass().getName().indexOf("Basic") < 0))
              {
                assertTrue(emptyReference.equals(bean));
                assertTrue(bean.equals(emptyReference));
              }
                assertEquals(emptyReference.hashCode(), bean.hashCode());
          }
      }
    
    private final static SimpleDateFormat EXIF_DATE_FORMAT = new SimpleDateFormat("yyyy:MM:dd HH:mm:ss");
    
    private Object createValue (final int i, final Class<?> propertyType, final String propertyName) 
      throws Exception 
      {
        Object value = null;

        if (i == ITERATIONS / 2)
          {
            return null;
          }       
        else if (propertyType.equals(String.class)) 
          {
            if (propertyName.startsWith("dateTime"))
              {
                final Date date = new Date(i, i, i);
                value = EXIF_DATE_FORMAT.format(date);
              }
            else
              {
                value = "Value " + i;
              }
          }
        else if (propertyType.equals(Integer.class)) 
          {
            value = new Integer(i);
          }
        else if (propertyType.equals(Long.class)) 
          {
            value = new Long(i);
          }
        else if (propertyType.equals(Float.class))
          {
            value = new Float(i);
          }
        else if (propertyType.equals(Double.class))
          {
            value = new Double(i);
          }
        else if (propertyType.equals(boolean.class)) 
          {
            value = (i % 2) == 0; // start the sequence with true
          } 
        else if (propertyType.equals(Date.class)) 
          {
            value = new Date(i, i, i);
          }
        else if (propertyType.equals(Dimension.class)) 
          {
            value = new Dimension(i, i);
          }
        else if (propertyType.getName().equals("it.tidalwave.image.Rational")) 
          {
            final Constructor<?> constructor = propertyType.getConstructor(int.class, int.class);
            value = constructor.newInstance((int)i, (int)i);
          }
        else if (propertyType.isArray() && propertyType.getComponentType().equals(byte.class))
          {
            value = new byte[] { (byte)i, (byte)(i+1), (byte)(i+2),(byte)(i+3) };  
          }
        else if (propertyType.isArray() && propertyType.getComponentType().equals(int.class))
          {
            value = new int[] { i, i+1, i+2, i+3 };  
          }
        else if (propertyType.isArray() && propertyType.getComponentType().equals(float.class))
          {
            value = new float[] { (float)i, (float)(i+1), (float)(i+2), (float)(i+3) };
          }
        else if (propertyType.isArray() && propertyType.getComponentType().equals(Float.class))
          {
            value = new Float[] { (float)i, (float)(i+1), (float)(i+2), (float)(i+3) };
          }
        else if (propertyType.isArray() && propertyType.getComponentType().equals(double.class))
          {
            value = new double[] { (double)i, (double)(i+1), (double)(i+2), (double)(i+3) };
          }
        else if (propertyType.isArray() && propertyType.getComponentType().getName().equals("it.tidalwave.image.Rational"))
          {
            final Constructor<?> constructor = propertyType.getComponentType().getConstructor(int.class, int.class);
            value = Array.newInstance(propertyType.getComponentType(), 4);
            Array.set(value, 0, constructor.newInstance(1, 1));
            Array.set(value, 1, constructor.newInstance(2, 2));
            Array.set(value, 2, constructor.newInstance(3, 3));
            Array.set(value, 3, constructor.newInstance(4, 4));
          }
        else if (propertyType.isEnum())
          {
            final Method method = propertyType.getMethod("values");
            final Object[] values = (Object[])method.invoke(null);
            return values[i % values.length];
          }
        else if (propertyType.equals(ICC_Profile.class))
          {
            final int[] colorSpaces = { ColorSpace.CS_GRAY, ColorSpace.CS_sRGB, ColorSpace.CS_LINEAR_RGB };
            return ICC_Profile.getInstance(colorSpaces[i % colorSpaces.length]);
          }
        
        assertNotNull("Unsupported type: " + propertyType, value);

        return value;
      }    
    
    protected static <T> T clone (final T bean)
      throws Exception
      {
        assertTrue("Not serializable: " + bean, bean instanceof Serializable);
       
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(bean);
        oos.close();
        final ByteArrayInputStream bais = new ByteArrayInputStream(baos.toByteArray());
        final ObjectInputStream ois = new ObjectInputStream(bais);
        final T result = (T)ois.readObject();
        ois.close();
        
        return result;
      }
    
    protected static String toStringx (final Object object)
      {
        if (object == null)
          {
            return "null";  
          }
        
        else if (object.getClass().isArray())
          {
            final int length = Array.getLength(object);
            final Object[] temp = new Object[length];
            copyArray(object, temp);
//            System.arraycopy(object, 0, temp, 0, temp.length);
            return Arrays.asList(temp).toString();  
          }
        else
          {
            return object.toString();
          }
      }
     
    public static void copyArray (final Object source, final Object dest) 
      {
        for (int i = 0; i < Array.getLength(source); i++) 
          {
            Array.set(dest, i, Array.get(source, i));
          }
      }
  }