/***********************************************************************************************************************
 *
 * blueMarine Metadata - open source media workflow
 * Copyright (C) 2007-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluemarine.tidalwave.it
 * SCM: https://kenai.com/hg/bluemarine~metadata-src
 *
 **********************************************************************************************************************/
package it.tidalwave.metadata.impl;

import it.tidalwave.metadata.spi.TestSupport;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import it.tidalwave.util.logging.Logger;
import java.io.IOException;
import org.openide.loaders.DataObject;
import it.tidalwave.metadata.Metadata;
import it.tidalwave.metadata.Metadata.ReplaceOption;
import it.tidalwave.metadata.Metadata.StorageType;
import it.tidalwave.metadata.MetadataItemHolder;
import it.tidalwave.metadata.NoSuchMetadataTypeException;
import static it.tidalwave.metadata.Metadata.StorageType.*;
import it.tidalwave.metadata.mock.MetadataItemHolderMock;
import it.tidalwave.metadata.mock.MockItem1;
import it.tidalwave.metadata.mock.MockItem2;
import it.tidalwave.metadata.mock.MetadataItemInterceptorMock1a;
import it.tidalwave.metadata.mock.MetadataItemInterceptorMock1b;
import it.tidalwave.metadata.mock.MetadataItemInterceptorMock2a;
import it.tidalwave.metadata.mock.MetadataItemProviderMock1;
import it.tidalwave.metadata.mock.MetadataItemProviderMock2;
import it.tidalwave.metadata.mock.MetadataItemSinkMock1a;
import it.tidalwave.metadata.mock.MetadataItemSourceMock1b;
import it.tidalwave.metadata.mock.MetadataItemSourceMock1a;
import it.tidalwave.metadata.mock.MetadataItemSourceMock1Internal;
import javax.swing.JLabel;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public class MetadataImplTest extends TestSupport 
  {
    private static final String CLASS = MetadataImplTest.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);
    
    private DataObject dataObject;
    private MetadataImpl metadata;
    
    @Before
    public void setupFixture() 
      throws IOException 
      {
        logger.info("******** setup()");
        dataObject = createDataObject("xxxx");
        metadata = new MetadataImpl(dataObject);
        MetadataItemInterceptorMock1a.metadataList.clear();
        MetadataItemInterceptorMock1a.holderList.clear();
        MetadataItemInterceptorMock1a.shouldThrowException = false;
        MetadataItemInterceptorMock1b.metadataList.clear();
        MetadataItemInterceptorMock1b.holderList.clear();
        MetadataItemInterceptorMock1b.shouldThrowException = false;
        MetadataItemInterceptorMock2a.metadataList.clear();
        MetadataItemInterceptorMock2a.holderList.clear();
        MetadataItemInterceptorMock2a.shouldThrowException = false;
        MetadataItemSinkMock1a.reset();
        logger.info("******** setup() done");
      }

    @After
    public void tearDownFixture() 
      {
        dataObject = null;
        metadata = null;
      }
    
    @Test
    public void testInitialization()
      {
        logger.info("******** testInitialization()");
        assertEquals(2, MetadataImpl.providerMapByClass.size());
        assertEquals(3, MetadataImpl.interceptors.size());
        assertTrue(MetadataImpl.providerMapByClass.get(MockItem1.class) instanceof MetadataItemProviderMock1);
        assertTrue(MetadataImpl.providerMapByClass.get(MockItem2.class) instanceof MetadataItemProviderMock2);
        assertTrue(MetadataImpl.interceptors.get(0) instanceof MetadataItemInterceptorMock1a);
        assertTrue(MetadataImpl.interceptors.get(1) instanceof MetadataItemInterceptorMock1b);
        assertTrue(MetadataImpl.interceptors.get(2) instanceof MetadataItemInterceptorMock2a);
      }
    
    @Test
    public void testGetItemClasses()
      {
        final ArrayList<Class<?>> itemClasses = new ArrayList<Class<?>>(metadata.getItemClasses());
        Collections.sort(itemClasses, new Comparator<Class<?>>() 
          { 
            public int compare (final Class<?> c1, final Class<?> c2) 
              {
                return c1.getName().compareTo(c2.getName());
              }
          });
        assertEquals(Arrays.asList(MockItem1.class, MockItem2.class), itemClasses);
      }
    
    @Test
    public void testLazyLoading()
      {
        logger.info("******** testLazyLoading()");
        assertTrue(metadata.holderMapByOption.isEmpty());
      }
    
    @Test
    public void testFindItemExternal()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testFindItemExternal()");
        final MetadataItemHolder<MockItem1> holder1 = metadata.findOrCreateItem(MockItem1.class, EXTERNAL);
        assertNotNull(holder1);
        final MetadataItemHolder<MockItem1> holder2 = metadata.findOrCreateItem(MockItem1.class, EXTERNAL);
        assertNotNull(holder2);
        assertTrue(holder1 == holder2);
        final MockItem1 item1 = holder1.get();
        final MockItem1 item2 = holder2.get();
        assertTrue(item1 == item2);
        assertEquals(MetadataItemSourceMock1b.DATE, holder1.getLatestModificationTime()); // most recent
      }
    
    @Test
    public void testFindItemsExternal()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testFindItemsExternal()");
        final List<MetadataItemHolder<MockItem1>> holders1 = metadata.findOrCreateItems(MockItem1.class, EXTERNAL);
        assertNotNull(holders1);
        assertEquals(2, holders1.size());
        final List<MetadataItemHolder<MockItem1>> holders2 = metadata.findOrCreateItems(MockItem1.class, EXTERNAL);
        assertEquals(2, holders2.size());
        assertNotNull(holders2);
        assertTrue(holders1.equals(holders2));
        final MetadataItemHolder<MockItem1> holder1 = holders1.get(0);
        final MetadataItemHolder<MockItem1> holder2 = holders1.get(1);
        assertEquals(MetadataItemSourceMock1b.DATE, holder1.getLatestModificationTime()); // most recent
        assertEquals(MetadataItemSourceMock1a.DATE, holder2.getLatestModificationTime());
      }
    
    @Test
    public void testFindItemsInternal()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testFindItemsInternal()");
        final List<MetadataItemHolder<MockItem1>> holders1 = metadata.findOrCreateItems(MockItem1.class, INTERNAL);
        assertNotNull(holders1);
        assertEquals(1, holders1.size());
        final List<MetadataItemHolder<MockItem1>> holders2 = metadata.findOrCreateItems(MockItem1.class, INTERNAL);
        assertEquals(1, holders2.size());
        assertNotNull(holders2);
        assertTrue(holders1.equals(holders2));
        final MetadataItemHolder<MockItem1> holder1 = holders1.get(0);
        assertEquals(MetadataItemSourceMock1Internal.DATE, holder1.getLatestModificationTime()); 
      }
    
    @Test
    public void testFindItemsAny()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testFindItemsAny()");
        final List<MetadataItemHolder<MockItem1>> items1 = metadata.findOrCreateItems(MockItem1.class, ANY_TYPE);
        assertNotNull(items1);
        assertEquals(items1.toString(), 3, items1.size());
        final List<MetadataItemHolder<MockItem1>> items2 = metadata.findOrCreateItems(MockItem1.class, ANY_TYPE);
        assertEquals(items1.toString(), 3, items2.size());
        assertNotNull(items2);
        assertTrue(String.format("%s vs %s", items1, items2), items1.equals(items2));
        final MetadataItemHolder<MockItem1> item1 = items1.get(0);
        final MetadataItemHolder<MockItem1> item2 = items1.get(1);
        final MetadataItemHolder<MockItem1> item3 = items1.get(2);
        // We are using the different dates to distinguish the three items
        final List<Date> dates = new ArrayList<Date>();
        dates.add(item1.getLatestModificationTime());
        dates.add(item2.getLatestModificationTime());
        dates.add(item3.getLatestModificationTime());
        assertTrue(dates.contains(MetadataItemSourceMock1Internal.DATE));
        assertTrue(dates.contains(MetadataItemSourceMock1a.DATE));
        assertTrue(dates.contains(MetadataItemSourceMock1b.DATE));
      }
    
    @Test
    public void testFindItemWithNotSupportedClass() 

    {
        logger.info("******** testFindItemWithNotSupportedClass()");
        try
          {
            metadata.findOrCreateItem(String.class, EXTERNAL);
            fail("NoSuchMetadataTypeException expected");
          }
        catch (NoSuchMetadataTypeException e)
          {
            // ok
          }
      }
    
    @Test
    public void testFindItemsWithNotSupportedClass() 

    {
        logger.info("******** testFindItemsWithNotSupportedClass()");
        try
          {
            metadata.findOrCreateItems(String.class, EXTERNAL);
            fail("NoSuchMetadataTypeException expected");
          }
        catch (NoSuchMetadataTypeException e)
          {
            // ok
          }
      }
    
    @Test
    public void testLastModificationTime()
      {
        logger.info("******** testLastModificationTime()");
        assertEquals(MetadataItemSourceMock1b.DATE, metadata.getLatestModificationTime());
      }
    
    @Test
    public void testLastModificationTimeIsUpdated() 
      throws InterruptedException, NoSuchMetadataTypeException
      {
        logger.info("******** testLastModificationTimeIsUpdated()");
        assertEquals(MetadataItemSourceMock1b.DATE, metadata.getLatestModificationTime());
        final MetadataItemHolder<MockItem1> item1 = metadata.findOrCreateItem(MockItem1.class, EXTERNAL);
        
        Thread.sleep(1000);
        Date now = new Date();
        item1.get().setName("name1");
        assertTrue(metadata.getLatestModificationTime().getTime() - now.getTime() < 50);
        
        Thread.sleep(1000);
        now = new Date();
        item1.get().setName("name2");
        assertTrue(metadata.getLatestModificationTime().getTime() - now.getTime() < 50);
        
        final MetadataItemHolder<MockItem2> item2 = metadata.findOrCreateItem(MockItem2.class, EXTERNAL);
        Thread.sleep(1000);
        now = new Date();
        item2.get().setName("name2");
        assertTrue(metadata.getLatestModificationTime().getTime() - now.getTime() < 50);
      }
    
    @Test
    public void testFindItemWithNullClass()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testFindItemWithNullClass()");
        try
          {
            metadata.findOrCreateItem(null, ANY_TYPE);
            fail("Exception Expected");
          }
        catch (IllegalArgumentException e)
          {
            // ok
          }
      }
    
    @Test
    public void testFindItemsWithNullClass()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testFindItemsWithNullClass()");
        try
          {
            metadata.findOrCreateItems(null, ANY_TYPE);
            fail("Exception Expected");
          }
        catch (IllegalArgumentException e)
          {
            // ok
          }
      }
    
    @Test
    public void testGetDataObject()
      {
        logger.info("******** testGetDataObject()");
        assertTrue(dataObject == metadata.getDataObject());
      }
    
    @Test
    public void testCreateWithNullDataObject()
      {
        logger.info("******** testCreateWithNullDataObject()");
        try
          {
            new MetadataImpl(null);
            fail("Exception Expected");
          }
        catch (IllegalArgumentException e)
          {
            // ok
          }
      }
    
    @Test
    public void testRobustInterceptors()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testRobustInterceptors()");
        MetadataItemInterceptorMock1a.shouldThrowException = true;
        MetadataItemInterceptorMock1a.reset();
        MetadataItemInterceptorMock1b.shouldThrowException = true;
        MetadataItemInterceptorMock1b.reset();
        final MetadataItemHolder<MockItem1> holder = metadata.findOrCreateItem(MockItem1.class, EXTERNAL); 
        assertNotNull(holder);
        assertInterceptorsCalled(metadata, holder);
      }
    
    @Test
    public void testStoreWithNullItem()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testStoreWithNullItem()");
        try
          {
            metadata.storeItem(null);
            fail("Exception Expected");
          }
        catch (IllegalArgumentException e)
          {
            // ok
          }
      }
    
    @Test
    public void testStoreWithNullOption()
      throws NoSuchMetadataTypeException
      {
        logger.info("******** testStoreWithNullOption()");
        try
          {
            // Use JLabel as an example of unsupported class since it supports PropertyChangeListener
            metadata.storeItem(new MetadataItemHolderMock<JLabel>(new JLabel()), StorageType.EXTERNAL, null);
            fail("Exception Expected");
          }
        catch (IllegalArgumentException e)
          {
            // ok
          }
      }
    
    @Test
    public void testStoreWithNotSupportedClass() 
      {
        logger.info("******** testStoreWithNotSupportedClass()");
        try
          {
            // Use JLabel as an example of unsupported class since it supports PropertyChangeListener
            metadata.storeItem(new MetadataItemHolderMock<JLabel>(new JLabel()), StorageType.EXTERNAL);
            fail("NoSuchMetadataTypeException expected");
          }
        catch (NoSuchMetadataTypeException e)
          {
            // ok
          }
      }
    
    
    // TODO: test store with EXTERNAL, INTERNAL, by name, etc...
    // TODO: test robust sinks
    
    @Test
    public void testTypicalImportPattern()
      throws NoSuchMetadataTypeException
      {
        final Collection<Class<?>> itemClasses = metadata.getItemClasses();
        
        for (final Class<?> itemClass : itemClasses)
          {
            final List<MetadataItemHolder<?>> holders = new ArrayList<MetadataItemHolder<?>>(metadata.findOrCreateItems(itemClass, StorageType.EXTERNAL));
            Collections.reverse(holders);
            
            for (final MetadataItemHolder<?> holder : holders)
              {
                final int n = metadata.storeItem(holder, StorageType.INTERNAL, ReplaceOption.REPLACE);
                assertEquals((holder.get() instanceof MockItem1) ? 1 : 0, n);
              }
          }
        
        assertEquals(2, MetadataItemSinkMock1a.dataObjects.size());
        assertEquals(2, MetadataItemSinkMock1a.holders.size());
        assertEquals(2, MetadataItemSinkMock1a.storeOptions.size());
        assertEquals(MetadataItemSourceMock1a.DATE, MetadataItemSinkMock1a.holders.get(0).getLatestModificationTime());
        assertEquals(MetadataItemSourceMock1b.DATE, MetadataItemSinkMock1a.holders.get(1).getLatestModificationTime()); // most recent last
      }
    
    @Test
    public void testToString()
      {
        logger.info("******** testToString()");
        assertEquals(String.format("MetadataImpl[%s]", dataObject), metadata.toString());
      }
    
    @Test
    public void testSourceType()
      {
        assertTrue(ANY_TYPE.includes(ANY_TYPE));
        assertTrue(ANY_TYPE.includes(EXTERNAL));
        assertTrue(ANY_TYPE.includes(INTERNAL));
        assertFalse(EXTERNAL.includes(ANY_TYPE));
        assertTrue(EXTERNAL.includes(EXTERNAL));
        assertFalse(EXTERNAL.includes(INTERNAL));
        assertFalse(INTERNAL.includes(ANY_TYPE));
        assertFalse(INTERNAL.includes(EXTERNAL));
        assertTrue(INTERNAL.includes(INTERNAL));
      }
    
    @Test
    public void testNamesFilter()
      {
        final Metadata.NameFilter nameFilter1 = new Metadata.NameFilter("name1", "name2", "name3");
        assertTrue(nameFilter1.matches("name1"));
        assertTrue(nameFilter1.matches("name2"));
        assertTrue(nameFilter1.matches("name3"));
        assertFalse(nameFilter1.matches("name4"));
        
        final Metadata.NameFilter nameFilter2 = new Metadata.NameFilter(Arrays.asList("name1", "name2", "name3"));
        assertTrue(nameFilter2.matches("name1"));
        assertTrue(nameFilter2.matches("name2"));
        assertTrue(nameFilter2.matches("name3"));
        assertFalse(nameFilter2.matches("name4"));
      }
    
    @Test
    public void testOrigin()
      {
        final Metadata.OriginFilter origin1 = new Metadata.OriginFilter("origin1");
        assertTrue(origin1.matches("origin1"));
        assertFalse(origin1.matches("origin2"));
        assertEquals(1, origin1.getNames().size());
        assertEquals("origin1", origin1.getNames().iterator().next());
        assertEquals("OriginFilter[origin1]", origin1.toString());
        
        final Metadata.OriginFilter origin2 = new Metadata.OriginFilter("origin2");
        assertFalse(origin2.matches("origin1"));
        assertTrue(origin2.matches("origin2"));
        assertEquals(1, origin2.getNames().size());
        assertEquals("origin2", origin2.getNames().iterator().next());
        assertEquals("OriginFilter[origin2]", origin2.toString());
      }
            
    private void assertInterceptorsCalled (final MetadataImpl metadata, final MetadataItemHolder<MockItem1> holder)
      {
        assertEquals(MetadataItemInterceptorMock1a.metadataList.toString(), 2, MetadataItemInterceptorMock1a.metadataList.size());
        assertEquals(MetadataItemInterceptorMock1a.holderList.toString(), 2, MetadataItemInterceptorMock1a.holderList.size());
        assertTrue(metadata == MetadataItemInterceptorMock1a.metadataList.get(0));
        assertTrue(metadata == MetadataItemInterceptorMock1a.metadataList.get(1));
        assertTrue(MetadataItemInterceptorMock1a.holderList.contains(holder));
        
        assertEquals(MetadataItemInterceptorMock1b.metadataList.toString(), 2, MetadataItemInterceptorMock1b.metadataList.size());
        assertEquals(MetadataItemInterceptorMock1b.holderList.toString(), 2, MetadataItemInterceptorMock1b.holderList.size());
        assertTrue(metadata == MetadataItemInterceptorMock1b.metadataList.get(0));
        assertTrue(metadata == MetadataItemInterceptorMock1b.metadataList.get(1));
        assertTrue(MetadataItemInterceptorMock1b.holderList.contains(holder));
      }
  }